@@ -13,10 +13,10 @@ from .exceptions import (
 from .extensions.permessage_deflate import ClientPerMessageDeflateFactory
 from .handshake import build_request, check_response
 from .headers import (
-    build_extension_list, build_protocol_list, parse_extension_list,
-    parse_protocol_list
+    build_basic_auth, build_extension_list, build_subprotocol_list,
+    parse_extension_list, parse_subprotocol_list
 )
-from .http import USER_AGENT, build_headers, read_response
+from .http import USER_AGENT, Headers, read_response
 from .protocol import WebSocketCommonProtocol
 from .uri import parse_uri
 
@@ -51,17 +51,14 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
         """
         self.path = path
-        self.request_headers = build_headers(headers)
-        self.raw_request_headers = headers
+        self.request_headers = headers
 
         # Since the path and headers only contain ASCII characters,
         # we can keep this simple.
-        request = ['GET {path} HTTP/1.1'.format(path=path)]
-        request.extend('{}: {}'.format(k, v) for k, v in headers)
-        request.append('\r\n')
-        request = '\r\n'.join(request).encode()
+        request = 'GET {path} HTTP/1.1\r\n'.format(path=path)
+        request += str(headers)
 
-        self.writer.write(request)
+        self.writer.write(request.encode())
 
     @asyncio.coroutine
     def read_http_response(self):
@@ -81,8 +78,7 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         except ValueError as exc:
             raise InvalidMessage("Malformed HTTP message") from exc
 
-        self.response_headers = build_headers(headers)
-        self.raw_response_headers = headers
+        self.response_headers = headers
 
         return status_code, self.response_headers
 
@@ -118,7 +114,7 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
         header_values = headers.get_all('Sec-WebSocket-Extensions')
 
-        if header_values is not None:
+        if header_values:
 
             if available_extensions is None:
                 raise InvalidHandshake("No extensions supported")
@@ -163,7 +159,7 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         """
         Handle the Sec-WebSocket-Protocol HTTP response header.
 
-        Check that it contains a supported subprotocol.
+        Check that it contains exactly one supported subprotocol.
 
         Return the selected subprotocol.
 
@@ -172,13 +168,13 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
         header_values = headers.get_all('Sec-WebSocket-Protocol')
 
-        if header_values is not None:
+        if header_values:
 
             if available_subprotocols is None:
                 raise InvalidHandshake("No subprotocols supported")
 
             parsed_header_values = sum([
-                parse_protocol_list(header_value)
+                parse_subprotocol_list(header_value)
                 for header_value in header_values
             ], [])
 
@@ -210,23 +206,29 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         subprotocols in order of decreasing preference.
 
         If provided, ``extra_headers`` sets additional HTTP request headers.
-        It must be a mapping or an iterable of (name, value) pairs.
+        It must be a :class:`~websockets.http.Headers` instance, a
+        :class:`~collections.abc.Mapping`, or an iterable of ``(name, value)``
+        pairs.
 
         Raise :exc:`~websockets.exceptions.InvalidHandshake` if the handshake
         fails.
 
         """
-        request_headers = []
-        set_header = lambda k, v: request_headers.append((k, v))
-        is_header_set = lambda k: k in dict(request_headers).keys()
+        request_headers = Headers()
 
         if wsuri.port == (443 if wsuri.secure else 80):     # pragma: no cover
-            set_header('Host', wsuri.host)
+            request_headers['Host'] = wsuri.host
         else:
-            set_header('Host', '{}:{}'.format(wsuri.host, wsuri.port))
+            request_headers['Host'] = '{}:{}'.format(wsuri.host, wsuri.port)
+
+        if wsuri.user_info:
+            request_headers['Authorization'] = build_basic_auth(
+                *wsuri.user_info)
 
         if origin is not None:
-            set_header('Origin', origin)
+            request_headers['Origin'] = origin
+
+        key = build_request(request_headers)
 
         if available_extensions is not None:
             extensions_header = build_extension_list([
@@ -236,33 +238,31 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
                 )
                 for extension_factory in available_extensions
             ])
-            set_header('Sec-WebSocket-Extensions', extensions_header)
+            request_headers['Sec-WebSocket-Extensions'] = extensions_header
 
         if available_subprotocols is not None:
-            protocol_header = build_protocol_list(available_subprotocols)
-            set_header('Sec-WebSocket-Protocol', protocol_header)
+            protocol_header = build_subprotocol_list(available_subprotocols)
+            request_headers['Sec-WebSocket-Protocol'] = protocol_header
 
         if extra_headers is not None:
-            if isinstance(extra_headers, collections.abc.Mapping):
+            if isinstance(extra_headers, Headers):
+                extra_headers = extra_headers.raw_items()
+            elif isinstance(extra_headers, collections.abc.Mapping):
                 extra_headers = extra_headers.items()
             for name, value in extra_headers:
-                set_header(name, value)
+                request_headers[name] = value
 
-        if not is_header_set('User-Agent'):
-            set_header('User-Agent', USER_AGENT)
-
-        key = build_request(set_header)
+        request_headers.setdefault('User-Agent', USER_AGENT)
 
         yield from self.write_http_request(
             wsuri.resource_name, request_headers)
 
         status_code, response_headers = yield from self.read_http_response()
-        get_header = lambda k: response_headers.get(k, '')
 
         if status_code != 101:
             raise InvalidStatusCode(status_code)
 
-        check_response(get_header, key)
+        check_response(response_headers, key)
 
         self.extensions = self.process_extensions(
             response_headers, available_extensions)
@@ -275,12 +275,13 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
 class Connect:
     """
-    This coroutine connects to a WebSocket server at a given ``uri``.
+    Connect to the WebSocket server at the given ``uri``.
 
-    It yields a :class:`WebSocketClientProtocol` which can then be used to
-    send and receive messages.
+    :func:`connect` returns an awaitable. Awaiting it yields an instance of
+    :class:`WebSocketClientProtocol` which can then be used to send and
+    receive messages.
 
-    On Python ≥ 3.5, :func:`connect` can be used as a asynchronous context
+    On Python ≥ 3.5.1, :func:`connect` can be used as a asynchronous context
     manager. In that case, the connection is closed when exiting the context.
 
     :func:`connect` is a wrapper around the event loop's
@@ -310,7 +311,9 @@ class Connect:
     * ``subprotocols`` is a list of supported subprotocols in order of
       decreasing preference
     * ``extra_headers`` sets additional HTTP request headers – it can be a
-      mapping or an iterable of (name, value) pairs
+      :class:`~websockets.http.Headers` instance, a
+      :class:`~collections.abc.Mapping`, or an iterable of ``(name, value)``
+      pairs
     * ``compression`` is a shortcut to configure compression extensions;
       by default it enables the "permessage-deflate" extension; set it to
       ``None`` to disable compression
@@ -382,14 +385,7 @@ class Connect:
             factory, host, port, **kwds)
 
     @asyncio.coroutine
-    def __aenter__(self):
-        return (yield from self)
-
-    @asyncio.coroutine
-    def __aexit__(self, exc_type, exc_value, traceback):
-        yield from self.ws_client.close()
-
-    def __await__(self):
+    def __iter__(self):                                     # pragma: no cover
         transport, protocol = yield from self._creating_connection
 
         try:
@@ -400,20 +396,25 @@ class Connect:
                 extra_headers=protocol.extra_headers,
             )
         except Exception:
-            yield from protocol.close_connection(after_handshake=False)
+            yield from protocol.fail_connection()
             raise
 
         self.ws_client = protocol
         return protocol
 
-    __iter__ = __await__
 
-
-if sys.version_info[:2] <= (3, 4):                          # pragma: no cover
+# We can't define __await__ on Python < 3.5.1 because asyncio.ensure_future
+# didn't accept arbitrary awaitables until Python 3.5.1. We don't define
+# __aenter__ and __aexit__ either on Python < 3.5.1 to keep things simple.
+if sys.version_info[:3] <= (3, 5, 0):                       # pragma: no cover
     @asyncio.coroutine
     def connect(*args, **kwds):
-        return Connect(*args, **kwds).__await__()
+        return Connect(*args, **kwds).__iter__()
     connect.__doc__ = Connect.__doc__
 
 else:
+    from .py35.client import __aenter__, __aexit__, __await__
+    Connect.__aenter__ = __aenter__
+    Connect.__aexit__ = __aexit__
+    Connect.__await__ = __await__
     connect = Connect
