@@ -8,12 +8,15 @@ frames as specified in `sections 4 to 8 of RFC 6455`_.
 
 import asyncio
 import asyncio.queues
+import binascii
 import codecs
 import collections
 import enum
 import logging
 import random
 import struct
+import sys
+import warnings
 
 from .compatibility import asyncio_ensure_future
 from .exceptions import (
@@ -28,6 +31,16 @@ __all__ = ['WebSocketCommonProtocol']
 logger = logging.getLogger(__name__)
 
 
+# On Python ≥ 3.7, silence a deprecation warning that we can't address before
+# dropping support for Python < 3.5.
+warnings.filterwarnings(
+    action='ignore',
+    message=r"'with \(yield from lock\)' is deprecated "
+            r"use 'async with lock' instead",
+    category=DeprecationWarning,
+)
+
+
 # A WebSocket connection goes through the following four states, in order:
 
 class State(enum.IntEnum):
@@ -58,9 +71,9 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
             await process(message)
 
     The iterator yields incoming messages. It exits normally when the
-    connection is closed with the status code 1000 OK. It raises a
-    :exc:`~websockets.exceptions.ConnectionClosed` exception when the
-    connection is closed with any other status code.
+    connection is closed with the status code 1000 (OK) or 1001 (going away).
+    It raises a :exc:`~websockets.exceptions.ConnectionClosed` exception when
+    the connection is closed with any other status code.
 
     The ``host``, ``port`` and ``secure`` parameters are simply stored as
     attributes for handlers that need them.
@@ -103,13 +116,12 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
     current implementation of ``FlowControlMixin``).
 
     As soon as the HTTP request and response in the opening handshake are
-    processed, the request path is available in the :attr:`path` attribute,
-    and the request and response HTTP headers are available:
+    processed:
 
-    * as a :class:`~http.client.HTTPMessage` in the :attr:`request_headers`
-      and :attr:`response_headers` attributes
-    * as an iterable of (name, value) pairs in the :attr:`raw_request_headers`
-      and :attr:`raw_response_headers` attributes
+    * the request path is available in the :attr:`path` attribute;
+    * the request and response HTTP headers are available in the
+      :attr:`request_headers` and :attr:`response_headers` attributes,
+      which are :class:`~websockets.http.Headers` instances.
 
     These attributes must be treated as immutable.
 
@@ -164,13 +176,12 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # Subclasses implement the opening handshake and, on success, execute
         # :meth:`connection_open()` to change the state to OPEN.
         self.state = State.CONNECTING
+        logger.debug("%s - state = CONNECTING", self.side)
 
         # HTTP protocol parameters.
         self.path = None
         self.request_headers = None
-        self.raw_request_headers = None
         self.response_headers = None
-        self.raw_response_headers = None
 
         # WebSocket protocol parameters.
         self.extensions = []
@@ -196,6 +207,9 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # Task running the data transfer.
         self.transfer_data_task = None
 
+        # Exception that occurred during data transfer, if any.
+        self.transfer_data_exc = None
+
         # Task closing the TCP connection.
         self.close_connection_task = None
 
@@ -221,6 +235,7 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # 4.1. The WebSocket Connection is Established.
         assert self.state is State.CONNECTING
         self.state = State.OPEN
+        logger.debug("%s - state = OPEN", self.side)
         # Start the task that receives incoming WebSocket messages.
         self.transfer_data_task = asyncio_ensure_future(
             self.transfer_data(), loop=self.loop)
@@ -268,7 +283,18 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         .. _EAFP: https://docs.python.org/3/glossary.html#term-eafp
 
         """
-        return self.state is State.OPEN
+        return self.state is State.OPEN and not self.transfer_data_task.done()
+
+    @property
+    def closed(self):
+        """
+        This property is ``True`` once the connection is closed.
+
+        Be aware that both :attr:`open` and :attr`closed` are ``False`` during
+        the opening and closing sequences.
+
+        """
+        return self.state is State.CLOSED
 
     @asyncio.coroutine
     def recv(self):
@@ -300,9 +326,10 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # Don't yield from self.ensure_open() here because messages could be
         # received before the closing frame even if the connection is closing.
 
-        # Wait for a message until the connection is closed
+        # Wait for a message until the connection is closed.
         next_message = asyncio_ensure_future(
             self.messages.get(), loop=self.loop)
+        # See https://bugs.python.org/issue23859 for cancellation handling.
         try:
             done, pending = yield from asyncio.wait(
                 [next_message, self.transfer_data_task],
@@ -312,15 +339,15 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
             next_message.cancel()
             raise
 
-        # Now there's no need to yield from self.ensure_open(). Either a
-        # message was received or the connection was closed.
-
         if next_message in done:
             return next_message.result()
         else:
             next_message.cancel()
             if not self.legacy_recv:
-                raise ConnectionClosed(self.close_code, self.close_reason)
+                assert self.state in [State.CLOSING, State.CLOSED]
+                # Wait until the connection is closed to raise
+                # ConnectionClosed with the correct code and reason.
+                yield from self.ensure_open()
 
     @asyncio.coroutine
     def send(self, data):
@@ -360,20 +387,15 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         ``code`` must be an :class:`int` and ``reason`` a :class:`str`.
 
         """
-        if self.state is State.OPEN:
-            # 7.1.2. Start the WebSocket Closing Handshake
-            # 7.1.3. The WebSocket Closing Handshake is Started
-            frame_data = serialize_close(code, reason)
-            try:
-                yield from asyncio.wait_for(
-                    self.write_frame(OP_CLOSE, frame_data),
-                    self.timeout, loop=self.loop)
-            except asyncio.TimeoutError:
-                # If the close frame cannot be sent because the send buffers
-                # are full, the closing handshake won't complete anyway.
-                # Cancel the data transfer task to shut down faster.
-                # Cancelling a task is idempotent.
-                self.transfer_data_task.cancel()
+        try:
+            yield from asyncio.wait_for(
+                self.write_close_frame(serialize_close(code, reason)),
+                self.timeout, loop=self.loop)
+        except asyncio.TimeoutError:
+            # If the close frame cannot be sent because the send buffers
+            # are full, the closing handshake won't complete anyway.
+            # Fail the connection to shut down faster.
+            self.fail_connection()
 
         # If no close frame is received within the timeout, wait_for() cancels
         # the data transfer task and raises TimeoutError. Then transfer_data()
@@ -388,8 +410,9 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
             # is cancelled before the timeout elapses (on Python ≥ 3.4.3).
             # This helps closing connections when shutting down a server.
             yield from asyncio.wait_for(
-                self.transfer_data_task, self.timeout, loop=self.loop)
-        except asyncio.TimeoutError:
+                self.transfer_data_task,
+                self.timeout, loop=self.loop)
+        except (asyncio.TimeoutError, asyncio.CancelledError):
             pass
 
         # Wait for the close connection task to close the TCP connection.
@@ -464,10 +487,18 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         """
         # Handle cases from most common to least common for performance.
         if self.state is State.OPEN:
-            return
+            # If self.transfer_data_task exited without a closing handshake,
+            # self.close_connection_task may be closing it, going straight
+            # from OPEN to CLOSED.
+            if self.transfer_data_task.done():
+                yield from asyncio.shield(self.close_connection_task)
+                raise ConnectionClosed(self.close_code, self.close_reason)
+            else:
+                return
 
         if self.state is State.CLOSED:
-            raise ConnectionClosed(self.close_code, self.close_reason)
+            raise ConnectionClosed(
+                self.close_code, self.close_reason) from self.transfer_data_exc
 
         if self.state is State.CLOSING:
             # If we started the closing handshake, wait for its completion to
@@ -477,7 +508,8 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
             # that case self.close_connection_task will complete even faster.
             if self.close_code is None:
                 yield from asyncio.shield(self.close_connection_task)
-            raise ConnectionClosed(self.close_code, self.close_reason)
+            raise ConnectionClosed(
+                self.close_code, self.close_reason) from self.transfer_data_exc
 
         # Control may only reach this point in buggy third-party subclasses.
         assert self.state is State.CONNECTING
@@ -498,20 +530,42 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
                 if msg is None:
                     break
                 yield from self.messages.put(msg)
-        except asyncio.CancelledError:
-            # This happens if self.close() cancels self.transfer_data_task.
-            pass
-        except WebSocketProtocolError:
-            yield from self.fail_connection(1002)
-        except asyncio.IncompleteReadError:
-            yield from self.fail_connection(1006)
-        except UnicodeDecodeError:
-            yield from self.fail_connection(1007)
-        except PayloadTooBig:
-            yield from self.fail_connection(1009)
-        except Exception:
-            logger.warning("Error in data transfer", exc_info=True)
-            yield from self.fail_connection(1011)
+
+        except asyncio.CancelledError as exc:
+            self.transfer_data_exc = exc
+            # If fail_connection() cancels this task, avoid logging the error
+            # twice and failing the connection again.
+            raise
+
+        except WebSocketProtocolError as exc:
+            self.transfer_data_exc = exc
+            self.fail_connection(1002)
+
+        except (ConnectionError, EOFError) as exc:
+            # Reading data with self.reader.readexactly may raise:
+            # - most subclasses of ConnectionError if the TCP connection
+            #   breaks, is reset, or is aborted;
+            # - IncompleteReadError, a subclass of EOFError, if fewer
+            #   bytes are available than requested.
+            self.transfer_data_exc = exc
+            self.fail_connection(1006)
+
+        except UnicodeDecodeError as exc:
+            self.transfer_data_exc = exc
+            self.fail_connection(1007)
+
+        except PayloadTooBig as exc:
+            self.transfer_data_exc = exc
+            self.fail_connection(1009)
+
+        except Exception as exc:
+            # This shouldn't happen often because exceptions expected under
+            # regular circumstances are handled above. If it does, consider
+            # catching and handling more exceptions.
+            logger.error("Error in data transfer", exc_info=True)
+
+            self.transfer_data_exc = exc
+            self.fail_connection(1011)
 
     @asyncio.coroutine
     def read_message(self):
@@ -592,28 +646,52 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
 
             # 5.5. Control Frames
             if frame.opcode == OP_CLOSE:
-                # Make sure the close frame is valid before echoing it.
-                code, reason = parse_close(frame.data)
                 # 7.1.5.  The WebSocket Connection Close Code
                 # 7.1.6.  The WebSocket Connection Close Reason
-                self.close_code, self.close_reason = code, reason
-                if self.state is State.OPEN:
-                    # 7.1.3. The WebSocket Closing Handshake is Started
-                    yield from self.write_frame(OP_CLOSE, frame.data)
+                self.close_code, self.close_reason = parse_close(frame.data)
+                # Echo the original data instead of re-serializing it with
+                # serialize_close() because that fails when the close frame is
+                # empty and parse_close() synthetizes a 1005 close code.
+                yield from self.write_close_frame(frame.data)
                 return
 
             elif frame.opcode == OP_PING:
                 # Answer pings.
+                # Replace by frame.data.hex() when dropping Python < 3.5.
+                ping_hex = binascii.hexlify(frame.data).decode() or '[empty]'
+                logger.debug("%s - received ping, sending pong: %s",
+                             self.side, ping_hex)
                 yield from self.pong(frame.data)
 
             elif frame.opcode == OP_PONG:
-                # Do not acknowledge pings on unsolicited pongs.
+                # Acknowledge pings on solicited pongs.
                 if frame.data in self.pings:
                     # Acknowledge all pings up to the one matching this pong.
                     ping_id = None
+                    ping_ids = []
                     while ping_id != frame.data:
                         ping_id, pong_waiter = self.pings.popitem(0)
+                        ping_ids.append(ping_id)
                         pong_waiter.set_result(None)
+                    pong_hex = (
+                        binascii.hexlify(frame.data).decode() or '[empty]')
+                    logger.debug("%s - received solicited pong: %s",
+                                 self.side, pong_hex)
+                    ping_ids = ping_ids[:-1]
+                    if ping_ids:
+                        pings_hex = ', '.join(
+                            binascii.hexlify(ping_id).decode() or '[empty]'
+                            for ping_id in ping_ids
+                        )
+                        plural = 's' if len(ping_ids) > 1 else ''
+                        logger.debug(
+                            "%s - acknowledged previous ping%s: %s",
+                            self.side, plural, pings_hex)
+                else:
+                    pong_hex = (
+                        binascii.hexlify(frame.data).decode() or '[empty]')
+                    logger.debug("%s - received unsolicited pong: %s",
+                                 self.side, pong_hex)
 
             # 5.6. Data Frames
             else:
@@ -635,17 +713,12 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         return frame
 
     @asyncio.coroutine
-    def write_frame(self, opcode, data=b''):
+    def write_frame(self, opcode, data=b'', _expected_state=State.OPEN):
         # Defensive assertion for protocol compliance.
-        if self.state is not State.OPEN:                    # pragma: no cover
+        if self.state is not _expected_state:               # pragma: no cover
             raise InvalidState("Cannot write to a WebSocket "
                                "in the {} state".format(self.state.name))
 
-        # Make sure no other frame will be sent after a close frame. Do this
-        # before yielding control to avoid sending more than one close frame.
-        if opcode == OP_CLOSE:
-            self.state = State.CLOSING
-
         frame = Frame(True, opcode, data)
         logger.debug("%s > %s", self.side, frame)
         frame.write(
@@ -669,9 +742,10 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
                 yield from self.writer.drain()
         except ConnectionError:
             # Terminate the connection if the socket died.
-            yield from self.fail_connection(1006)
-            # And raise an exception, since the frame couldn't be sent.
-            raise ConnectionClosed(self.close_code, self.close_reason)
+            self.fail_connection()
+            # Wait until the connection is closed to raise ConnectionClosed
+            # with the correct code and reason.
+            yield from self.ensure_open()
 
     def writer_is_closing(self):
         """
@@ -693,7 +767,26 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
                 return transport._closed
 
     @asyncio.coroutine
-    def close_connection(self, after_handshake=True):
+    def write_close_frame(self, data=b''):
+        """
+        Write a close frame if and only if the connection state is OPEN.
+
+        This dedicated coroutine must be used for writing close frames to
+        ensure that at most one close frame is sent on a given connection.
+
+        """
+        # Test and set the connection state before sending the close frame to
+        # avoid sending two frames in case of concurrent calls.
+        if self.state is State.OPEN:
+            # 7.1.3. The WebSocket Closing Handshake is Started
+            self.state = State.CLOSING
+            logger.debug("%s - state = CLOSING", self.side)
+
+            # 7.1.2. Start the WebSocket Closing Handshake
+            yield from self.write_frame(OP_CLOSE, data, State.CLOSING)
+
+    @asyncio.coroutine
+    def close_connection(self):
         """
         7.1.1. Close the WebSocket Connection
 
@@ -701,17 +794,33 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         this coroutine in a task. It waits for the data transfer phase to
         complete then it closes the TCP connection cleanly.
 
-        When the opening handshake fails, the client or the server runs this
-        coroutine with ``after_handshake=False`` to close the TCP connection.
+        When the opening handshake fails, :meth:`fail_connection` does the
+        same. There's no data transfer phase in that case.
 
         """
         try:
             # Wait for the data transfer phase to complete.
-            if after_handshake:
-                yield from self.transfer_data_task
+            if self.transfer_data_task is not None:
+                try:
+                    yield from self.transfer_data_task
+                except asyncio.CancelledError:
+                    pass
 
-            # A client should wait for a TCP Close from the server.
-            if self.is_client and after_handshake:
+            # Cancel all pending pings because they'll never receive a pong.
+            for ping in self.pings.values():
+                ping.cancel()
+            if self.pings:
+                pings_hex = ', '.join(
+                    binascii.hexlify(ping_id).decode() or '[empty]'
+                    for ping_id in self.pings
+                )
+                plural = 's' if len(self.pings) > 1 else ''
+                logger.debug(
+                    "%s - cancelled pending ping%s: %s",
+                    self.side, plural, pings_hex)
+
+            # A client should wait for a TCP close from the server.
+            if self.is_client and self.transfer_data_task is not None:
                 if (yield from self.wait_for_connection_lost()):
                     return
                 logger.debug(
@@ -732,8 +841,11 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
             # The try/finally ensures that the transport never remains open,
             # even if this coroutine is cancelled (for example).
 
-            # Closing a transport is idempotent. If the transport was already
-            # closed, for example from eof_received(), it's fine.
+            # If connection_lost() was called, the TCP connection is closed.
+            # However, if TLS is enabled, the transport still needs closing.
+            # Else asyncio complains: ResourceWarning: unclosed transport.
+            if self.connection_lost_waiter.done() and not self.secure:
+                return
 
             # Close the TCP connection. Buffers are flushed asynchronously.
             logger.debug(
@@ -773,20 +885,72 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # and the moment this coroutine resumes running.
         return self.connection_lost_waiter.done()
 
-    @asyncio.coroutine
-    def fail_connection(self, code=1011, reason=''):
+    def fail_connection(self, code=1006, reason=''):
         """
         7.1.7. Fail the WebSocket Connection
 
+        This requires:
+
+        1. Stopping all processing of incoming data, which means cancelling
+           :attr:`transfer_data_task`. The close code will be 1006 unless a
+           close frame was received earlier.
+
+        2. Sending a close frame with an appropriate code if the opening
+           handshake succeeded and the other side is likely to process it.
+
+        3. Closing the connection. :meth:`close_connection` takes care of
+           this once :attr:`transfer_data_task` exits after being cancelled.
+
+        (The specification describes these steps in the opposite order.)
+
+        Return a :class:`~asyncio.Task` that completes when the TCP connection
+        is closed.
+
         """
         logger.debug(
             "%s ! failing WebSocket connection: %d %s",
             self.side, code, reason,
         )
+
+        # Cancel transfer_data_task if the opening handshake succeeded.
+        # cancel() is idempotent and ignored if the task is done already.
+        if self.transfer_data_task is not None:
+            self.transfer_data_task.cancel()
+
+        # Send a close frame when the state is OPEN (a close frame was already
+        # sent if it's CLOSING), except when failing the connection because of
+        # an error reading from or writing to the network.
         # Don't send a close frame if the connection is broken.
-        if self.state is State.OPEN and code != 1006:
+        if code != 1006 and self.state is State.OPEN:
+
             frame_data = serialize_close(code, reason)
-            yield from self.write_frame(OP_CLOSE, frame_data)
+
+            # Write the close frame without draining the write buffer.
+
+            # Keeping fail_connection() synchronous guarantees it can't
+            # get stuck and simplifies the implementation of the callers.
+            # Not drainig the write buffer is acceptable in this context.
+
+            # This duplicates a few lines of code from write_close_frame()
+            # and write_frame().
+
+            self.state = State.CLOSING
+            logger.debug("%s - state = CLOSING", self.side)
+
+            frame = Frame(True, OP_CLOSE, frame_data)
+            logger.debug("%s > %s", self.side, frame)
+            frame.write(
+                self.writer.write,
+                mask=self.is_client,
+                extensions=self.extensions,
+            )
+
+        # Start close_connection_task if the opening handshake didn't succeed.
+        if self.close_connection_task is None:
+            self.close_connection_task = asyncio_ensure_future(
+                self.close_connection(), loop=self.loop)
+
+        return self.close_connection_task
 
     # asyncio.StreamReaderProtocol methods
 
@@ -804,7 +968,7 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         which means it's the best point for configuring it.
 
         """
-        logger.debug("%s - connection_made(%s)", self.side, transport)
+        logger.debug("%s - event = connection_made(%s)", self.side, transport)
         transport.set_write_buffer_limits(self.write_limit)
         super().connection_made(transport)
 
@@ -826,12 +990,12 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
            returned to have the same behavior on TLS and plain connections.
 
         3. The websockets protocol has its own closing handshake. Endpoints
-           close the TCP connection after sending a Close frame.
+           close the TCP connection after sending a close frame.
 
         As a consequence we revert to the previous, more useful behavior.
 
         """
-        logger.debug("%s - eof_received()", self.side)
+        logger.debug("%s - event = eof_received()", self.side)
         super().eof_received()
         return
 
@@ -840,10 +1004,13 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         7.1.4. The WebSocket Connection is Closed.
 
         """
-        logger.debug("%s - connection_lost(%s)", self.side, exc)
+        logger.debug("%s - event = connection_lost(%s)", self.side, exc)
         self.state = State.CLOSED
+        logger.debug("%s - state = CLOSED", self.side)
         if self.close_code is None:
             self.close_code = 1006
+        logger.debug("%s x code = %d, reason = %s", self.side,
+                     self.close_code, self.close_reason or '[empty]')
         # If self.connection_lost_waiter isn't pending, that's a bug, because:
         # - it's set only here in connection_lost() which is called only once;
         # - it must never be cancelled.
@@ -851,9 +1018,6 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         super().connection_lost(exc)
 
 
-try:
+if sys.version_info[:2] >= (3, 6):                          # pragma: no cover
     from .py36.protocol import __aiter__
-except (SyntaxError, ImportError):                          # pragma: no cover
-    pass
-else:
     WebSocketCommonProtocol.__aiter__ = __aiter__
